/*
    CDROM Low level functions
  ================================================================
  This code is done for making an easy conversion to plugin format
  ================================================================
*/
#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "winioctl.h"  // From the Win32 SDK \Mstools\Include
#include "ntddcdrm.h"  // From the Windows NT DDK \Ddk\Src\Storage\Inc

#include "type.h"
#include "Win32Def.h"
#include "Cd.h"

#define INT2BCD(a)  (((a)/10)*16+((a)%10))
#define BCD2INT(a)  (((a)>>4)*10+((a)&0xf))

#define RAW_SECTOR_SIZE     2352
#define COOKED_SECTOR_SIZE  2048

static HANDLE          hCD        = NULL;
static LPBYTE          lpSector   = NULL;
static int             CdError    = 0;
static UINT32          LastSector = -1;

#define MSF2SECT(m,s,f)  ( ( (m)*60 + (s) -2 )*75 + (f) )

// #define DEBUG_IOCTRL

// return NULL if something was wrong
UINT8 *CD_Read(UINT8 *param)
{
    DWORD   dwNotUsed;
    int     Ret;

    Ret = MSF2SECT(param[0], param[1], param[2]);
    if (Ret != LastSector)
    {
        SetFilePointer (hCD, 2048*Ret, NULL, FILE_BEGIN);
        LastSector = Ret;
    }
    LastSector++;

    Ret = ReadFile(hCD, lpSector+12, 2048, &dwNotUsed, NULL);

#ifdef DEBUG_IOCTRL
    printf("Ret = %d - nu = %d\n",Ret,dwNotUsed);
    if (!Ret) printf("Error = %08x\n",GetLastError());
#endif

    CdError = 0;
//    if (!Ret) CdError = 1;

// emulate header
    lpSector[0] = INT2BCD(param[0]);
    lpSector[1] = INT2BCD(param[1]);
    lpSector[2] = INT2BCD(param[2]);

    return lpSector;
}

// stop playing CDA
int CD_Stop(void)
{
    return FPSE_OK;
}

// Play CDA at given loc
int CD_Play(UINT8 *param)
{
    return FPSE_OK;
}

// Return Track specs
int CD_GetTD(UINT8 *result,int track)
{
    CDROM_TOC toc;
    DWORD     dwNotUsed;
    int       x;

    DeviceIoControl(hCD, IOCTL_CDROM_READ_TOC,
                    NULL, 0, &toc, CDROM_TOC_SIZE,
                    &dwNotUsed, NULL);

    for (x=0; x < MAXIMUM_NUMBER_TRACKS; x++)
    {
        if (toc.TrackData[x].TrackNumber == track)
        {

            result[1] = BCD2INT(toc.TrackData[x].Address[1]);
            result[2] = BCD2INT(toc.TrackData[x].Address[2]);
            result[3] = BCD2INT(toc.TrackData[x].Address[3]);
            return FPSE_OK;
        }
    }

    return FPSE_ERR;
}

// Get track number
int CD_GetTN(UINT8 *result)
{
    CDROM_TOC   toc;
    DWORD       dwNotUsed;

    DeviceIoControl(hCD, IOCTL_CDROM_READ_TOC,
                    NULL, 0, &toc, CDROM_TOC_SIZE,
                    &dwNotUsed, NULL);

    result[1] = toc.FirstTrack;
    result[2] = toc.LastTrack;

    return FPSE_OK;
}

// init the CD hardware
int CD_Open(UINT32 *par)
{
    FPSEWin32  *info=(FPSEWin32 *)(par);
    char        Drive[8]="\\\\.\\?:";
    char        Letter[4];

#ifndef DEBUG_IOCTRL
    info->ReadCfg("CDIOCTRL","DriveLetter",Letter);
    Drive[4] = Letter[0];
#else
    Drive[4] = 'E';
#endif

    hCD = CreateFile (Drive, GENERIC_READ,
                     FILE_SHARE_READ|FILE_SHARE_WRITE,
                     NULL, OPEN_EXISTING,
                     FILE_FLAG_NO_BUFFERING,
                     NULL);

    if (hCD == INVALID_HANDLE_VALUE) return FPSE_ERR;

    lpSector = VirtualAlloc (NULL, 4096,
                             MEM_COMMIT|MEM_RESERVE,
                             PAGE_READWRITE);

    CdError = 0;

    return FPSE_OK;
}

// shutdown...
void CD_Close(void)
{
    CD_Wait();

    if (lpSector != NULL)
        VirtualFree (lpSector, 0, MEM_RELEASE);
    lpSector = NULL;

    CloseHandle (hCD);
}

// check end operation (with it busmaster is great!)
int CD_Wait(void)
{
    if (CdError)
    {
        CdError = 0;
        return FPSE_ERR;
    }

    return FPSE_OK;
}

// Return the seek position during CDA play
void CD_GetSeek(UINT8 *par)
{

}

#ifdef DEBUG_IOCTRL
static void dump(UINT8 *data, int len)
{
    int   i,j;
    UINT8 c;

    for(i=0;i<len;i++)
    {
        if ((i&15)==0) printf("%04x:",i & ~15);
        printf("%02x ",data[i]);
        if ((i&15)==15) {
            for (j=0; j<16; j++)
            {
                c = data[(i & ~15)+j];
                if (c > 31 && c < 128) printf("%c",c);
                                  else printf(".");
            }
            printf("\n");
        }
    }

    printf("\n");
}

int main(int argc,char **argv)
{
    UINT8 param[8];
    int   i,res;
    char *buf;

    printf("FPSE CDCTRL TEST\n\nInit...\n");
    res = CD_Open(NULL);
    if (res != FPSE_OK) return 1;
    printf("Start reading...\n\n");

    for (i=0; i<23; i++)
    {
        param[0] = 0;
        param[1] = 2;
        param[2] = i;

        memset(lpSector,0,4096);

        buf = CD_Read(param);
        res = CD_Wait();
        if (res == FPSE_OK)
        {
            printf("frame %d:\n",i);
            dump(buf,RAW_SECTOR_SIZE);
        } else {
            printf("frame %d - read error\n",i);
        }
    }

    CD_Close();

    return 0;
}
#endif
